import java.awt.Dimension;
import java.sql.SQLException;
import java.util.*;

import jas.engine.*;
import jas.events.SimGroupEvent;
import jas.graph.*;
import jas.graphics.plot.TimeSeriesPlotter;
import jas.io.ParameterFrame;

import jas.statistics.db.CollectionTable;
import jas.statistics.db.Database;
import jas.statistics.db.Table;

/**
 *
 * <p>Title: JAS</p>
 * <p>Description: Java Agent-based Simulation library</p>
 * <p>Copyright (C) 2002 Michele Sonnessa</p>
 *
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * @author Michele Sonnessa
 *
 */
public class BlueNetModel extends SimModel //implements IGraphMLCreator
{
	//Model parameters
	public int 	numberOfAgents;
	public String fileSource = "";
	public String databaseName = "";
	public int numberOfGatePerMaster = 1;
	public int waitBeforeInquiry = 100;
	public boolean enablePrice = false;
	public int maxGenerationTime = 200;

	private Database database;
	private double maxHop = 0;

	private GraphViewer graphViewer;
	private String dataFile;
	private int generation = 0;

	//Simulation agents
	public static int agentCounter = 0;
	public ArrayList agentList;

	//Technical
	private ParameterFrame parameterFrame;

	public TimeSeriesPlotter plotter;

	public DeviceMap graph;
	public int xSize, ySize;

	public void setParameters()
	{
		parameterFrame = new ParameterFrame(this, "Bluetooth connector", path + "BlueNetParams.pForm.xml");
		addSimWindow(parameterFrame);
	}

	public void shuffleList()
	{
		Collections.shuffle(agentList);
	}

	public void simulationEnd()
	{
//		GraphML writer = new GraphML();
//		writer.save(graph, path + "graph.xml");
		System.out.println("Completed at time " + Sim.getAbsoluteTime() + " max hop = " + maxHop);
	}

	public void buildModel()
	{
		if (fileSource.length() > 0)
		{
			//loadGraph();
			buildGraph();
		}
		else
		{
			graph = new DeviceMap(xSize, ySize, numberOfGatePerMaster);
			agentList = new ArrayList();
			newNode();
			buildGraph();
		}

//		for (int i = 0; i < 5; i++) {
//			newNode();
//		}
//		newFile();


		if (databaseName.length() > 0)
		try {
			database = new Database(path, databaseName, false);

			CollectionTable table = database.addCollectionTable("Colors", Table.PK_SIMULATION_TIME, agentList);
			table.addIntColumn("CurrentColor");

			database.openDatabase();
		} catch (SQLException e) {
			e.printStackTrace();
		}

		plotter = new TimeSeriesPlotter("Service quality");
		plotter.addSeries("Avg comm failures", this, "getAvgCommunicationFailures", true);
		plotter.addSeries("Avg hops", this, "getAvgHops", true);
		addSimWindow(plotter);

		buildSchedule();
	}

//	private void loadGraph(){
//		GraphML f = new GraphML(this);
//		agentList = new ArrayList();
//		graph = (DirectedWeightedMultigraph) f.load(this, path + fileSource);
//		System.out.println("Graph read");
//	}

	private void buildGraph()
	{
		graphViewer = new GraphViewer();
		graphViewer.setGraph(graph, new Dimension(xSize, ySize));
		addSimWindow(graphViewer);
	}

	public void killNode()
	{
		Device dev = (Device) agentList.get(Sim.getRnd().getIntFromTo(0, agentList.size() - 1));
		graph.removeVertex(dev);
		agentList.remove(dev);
		return;
	}

	public void newNode()
	{
		Device agent = new Device(graph, agentCounter++, waitBeforeInquiry, enablePrice);
		agentList.add(agent);
		agent.setLabel("N"+(agentCounter));
		graph.placeDeviceRandomly(agent);
	}

	public void newFile()
	{
			Device src = (Device) agentList.get(Sim.getRnd().getIntFromTo(0, agentList.size() - 1));
			Device dest = (Device) agentList.get(Sim.getRnd().getIntFromTo(0, agentList.size() - 1));
			if (src == dest)
				return;

			int packets = Sim.getRnd().getIntFromTo(10, 1000);
			FileTransfer file = new FileTransfer(dest, packets);
			src.setFileTransfer(file);
	}

	private void buildSchedule()
	{
		SimGroupEvent g = eventList.scheduleGroup(0, 1);
		g.addCollectionEvent(agentList, Device.PROTOCOL_ACTIVITY);
		g.addCollectionEvent(agentList, Device.CHECK_CONNECTION);
		g.addCollectionEvent(agentList, Device.COMMUNICATE);
		g.addEvent(graphViewer, Sim.EVENT_UPDATE);
		if (database != null)
			g.addEvent(database, Sim.EVENT_UPDATE);
		g.addEvent(plotter, Sim.EVENT_UPDATE);

		for (int i = 0; i < maxGenerationTime; i += 5) {
			eventList.scheduleSimple(i, this, "newNode");
			//eventList.scheduleSimple(i + 2, this, "newFile");
		}

		for (int i = 0; i < maxGenerationTime; i += 20) {
			//eventList.scheduleSimple(i, this, "newNode");
			eventList.scheduleSimple(i + 2, this, "newFile");
		}

		//eventList.scheduleSystem(10001, Sim.EVENT_SIMULATION_END);
	}

	public double getAvgCommunicationFailures()
	{
		int connections = 0, successful = 0;
		for (Iterator iter = agentList.iterator(); iter.hasNext();) {
			Device element = (Device) iter.next();
			if (element.getFileTransfer() != null)
			{
				int fail = element.getFileTransfer().getFailed();
				if (fail > 0)
				{
					connections++;
					successful += fail;
				}
			}
		}

		double ratio = 0.0;

		if (connections > 0)
			ratio = successful / connections;

		if (ratio < 0)
			ratio = 0.0;
		if (ratio > 20)
			ratio = 20.0;

		return ratio;
	}

	public double getAvgHops()
	{
		int connections = 0, successful = 0;
		for (Iterator iter = agentList.iterator(); iter.hasNext();) {
			Device element = (Device) iter.next();
			if (element.getFileTransfer() != null)
			{
				int hop = element.getFileTransfer().getHops();
				if (hop > 0)
				{
					connections++;
					successful += hop;
				}
			}
		}

		double ratio = 0.0;

		if (connections > 0)
			ratio = successful / connections;

		if (ratio < 0)
			ratio = 0.0;

		if (ratio > maxHop)
			maxHop = ratio;

		if (Sim.getAbsoluteTime() > 100 && ratio == 0.0)
			eventList.scheduleSystem(Sim.getAbsoluteTime() + 1, Sim.EVENT_SIMULATION_END);

		return ratio;
	}

//	/* (non-Javadoc)
//	 * @see jas.graph.IGraphMLCreator#createNewNode(java.util.Map)
//	 */
//	public IRelationalAgent createNewNode(int id, Map attributes) {
//		Device agent = new Device(graph, id, commNoiseProbability);
//		agentList.add(agent);
//		agent.setLabel("V" + id);
//		agent.setColorPref(Integer.parseInt(attributes.get("colorPref").toString()));
//		agent.setColorMap(colorMap);
//		agent.setNodeColor();
//		return agent;
//	}

	/* (non-Javadoc)
	 * @see jas.graph.IGraphMLCreator#getNewGraph(org._3pq.jgrapht.Graph)
	 */
//	public void getNewGraph(Graph newGraph) {
//		graph = (DirectedWeightedMultigraph) newGraph;
//	}


}

